(**************************************************************************)
(*                     Deterministic Finite Automata                      *)
(**************************************************************************)

Implicit Arguments On.

Require FiniteSets.


(* Given an alphabet A and a set St,
   a deterministic finite automaton (DFA) is a 5-tuple (I,Q,q0,F,delta) where

      i) Q is a finite nonempty set of states (over St)
     ii) I is a finite nonempty set of inputs (over A)
    iii) delta is the transition function, mapping Q x I to Q
     iv) q0 is the initial state (member of Q)
      v) F is the set of final states (subset of Q).  	*)

Inductive df_automaton [ A:Set; St:Set; I:(fset A); Q:(fset St); q0:St;
      	       	         F:(fset St); delta:(mapping (product Q I) Q) ] : Prop
 :=
  df_automaton_intro :
      (member q0 Q) ->
      (subset F Q) ->
      (df_automaton A St I Q q0 F delta).


(**************************************************************************)
(*                       The set of accepted words                        *)
(**************************************************************************)

Require Words.
Require Sets_of_words.

(* First we define the notion of path: 	*)

Inductive df_path [ A:Set; St:Set; I:(fset A); Q:(fset St); 
      	       	    delta:(mapping (product Q I) Q) ] 
                  : St -> St -> (word A) -> Prop :=

  df_path_nil  : (e:St)(member e Q) -> (df_path A St I Q delta e e (nil A))

| df_path_cons : (e1,e2:St)(w:(word A))(a:A)
      	         (p:(member (e1,a) (product Q I)))
      	         (df_path A St I Q delta (map_app (delta (e1,a) p)) e2 w) ->
	         (df_path A St I Q delta e1 e2 (cons A a w)).


(* Property : there is a path in an automaton A from the state e1 to
   the state e2 for the word w iff (extw delta w e1)=e2. *)

Axiom df_path_inv : 
  (A,St:Set)(I:(fset A))(Q:(fset St))(delta:(mapping (product Q I) Q))
  (w:(word A))(Hw:(monoid I w))
  (e1,e2:St)(He1:(member e1 Q))
  (df_path delta e1 e2 w) <-> (map_app (extw delta Hw He1))=e2.

(**** TODO
Proof.
Intros A St I Q delta.
Induction w.
Intro Hw.
Split.
Intros Hpath.
Inversion Hpath.
Unfold extw.
Simpl.
Rewrite <- H1.
Auto.

Unfold extw.
Simpl.
Intros eq.
Rewrite -> eq.
(Apply df_path_nil; Auto).
Rewrite <- eq.
Assumption.

Generalize (monoid_inv_cons_w Hw). Intro Hw'.
Generalize (monoid_inv_cons_a Hw). Intro Ha.
Generalize (member_pair_prod He1 (monoid_inv_cons_a Hw)). Intro He1a.
Generalize (proj1_sig St [y:St](member y Q) (delta (e1,a) He1a)). Intro q'.
Generalize (sig_sig (delta (e1,a) He1a)). Intro Hq'.
Elim (HR Hw' (proj1_sig St [y:St](member y Q) (delta (e1,a) He1a))  e2 Hq').
Intros HRw'_left HRw'_right.
Generalize (HRw'_left H2).

Generalize (proj1_sig St [y:St](member y Q) (delta (e1,a) p)). Intro q'.
Generalize (sig_sig (delta (e1,a) p)). Intro Hq'.
Elim (HR Hw' (proj1_sig St [y:St](member y Q) (delta (e1,a) p))  e2 Hq').
Intros HRw'_left HRw'_right.
Generalize (HRw'_left H2).
Intro equ. Rewrite <- equ.
Intro.

Save.
***)


(* A finite automaton A accepts a word w if A, starting in q0, goes through
   some computation and reads all of w, and stops in some final state.	
   
   The set of all accepted words, T(A), is (df_accepts A St I Q q0 F delta). *)

Inductive df_accepts [ A:Set; St:Set; I:(fset A); Q:(fset St); q0:St;
      	       	       F:(fset St); delta:(mapping (product Q I) Q);
                       w:(word A)] : Prop :=

  df_accepts_intro : (f:St) (member f F) ->
      	       	     (df_path delta q0 f w) ->
		     (df_accepts A St I Q q0 F delta w).


Require Sets_of_words.

(* A set of words L is a regular set if there is some finite automaton
   A such that L=T(A). 
   regularset is defined in sort Set because we want to extract
   the finite automaton that recognized a language L when having
   done a proof that L is regular. *)

Inductive regularset [ A:Set; St:Set; I:(fset A); L:(wordset A) ] : Set :=
  regular_intro :
      (Q:(fset St))(q0:St)(F:(fset St))
      (delta:(mapping (product Q I) Q))
      (df_automaton q0 F delta) ->
      (eq_wordset L (df_accepts q0 F delta)) ->
      (regularset A St I L).
