(**************************************************************************)
(*                    Nondeterministic Finite Automata                    *)
(**************************************************************************)

Implicit Arguments On.

Require FiniteSets.

(* Given an alphabet A and a set St,
   a nondeterministic finite automaton (NFA) is a 5-tuple
   (I,Q,q0,F,delta) where

      i) Q is a finite nonempty set of states (over St)
     ii) I is a finite nonempty set of inputs (over A)
    iii) delta is a map from Q x I to 2^Q (power set of Q)
         Is is equivalent to consider delta as a subset of Q x I x Q.
     iv) q0 is the initial state (member of Q)
      v) F is the set of final states (subset of Q).  	*)

Inductive nf_automaton [ A:Set; St:Set; I:(fset A); Q:(fset St); q0:St;
      	       	         F:(fset St); delta:(fset St*(A*St)) ] : Prop :=
  nf_automaton_intro :
      (member q0 Q) ->
      (subset F Q) ->
      (subset delta (product Q (product I Q))) ->
      (nf_automaton A St I Q q0 F delta).

(**************************************************************************)
(*                       The set of accepted words                        *)
(**************************************************************************)

Require Words.

(* First we define the notion of path: there is a path in an automaton A
   from the state e1 to the state e2 for the word w iff delta^(e1,w)=e2. *)

Inductive nf_path [ A:Set; St:Set; I:(fset A); Q:(fset St);
      	       	    delta:(fset St*(A*St)) ] : St -> St -> (word A) -> Prop :=

  nf_path_nil  : (e:St)(member e Q) -> (nf_path A St I Q delta e e (nil A))

| nf_path_cons : (e1,e2,e:St)(w:(word A))(a:A)
      	         (nf_path A St I Q delta e1 e2 w) ->
	         (member a I) ->
	         (member (e,(a,e1)) delta) ->
	         (member e Q) ->
	         (nf_path A St I Q delta e e2 (cons A a w)).

(* A finite automaton A accepts a word w if A, starting in q0, goes through
   some computation and reads all of w, and stops in some final state.	
   
   The set of all accepted words, T(A), is (nf_accepts A St I Q q0 F delta). *)

Inductive nf_accepts [ A:Set; St:Set; I:(fset A); Q:(fset St); q0:St;
      	       	       F:(fset St); delta:(fset St*(A*St)); w:(word A)] : Prop
  :=
  nf_accepts_intro : (f:St) (member f F) ->
      	       	     (nf_path I Q delta q0 f w) ->
		     (nf_accepts A St I Q q0 F delta w).

Require Sets_of_words.

(* A set of words L is called nf_regular is there is some
   nondeterministic finite automaton A such that L=T(A).
   We will prove later that it corresponds to the notion of
   regular sets. *)

Inductive nf_regularset [ A:Set; St:Set; I:(fset A); L:(wordset A) ] : Set :=

  nf_regular_intro :
      (Q:(fset St))(q0:St)(F:(fset St))
      (delta:(fset St*(A*St)))
      (nf_automaton I Q q0 F delta) ->
      (eq_wordset L (nf_accepts I Q q0 F delta)) ->
      (nf_regularset A St I L).

(* To do the proof of Kleene's theorem, it is convenient to introduce
   a partiular case of NFA : those having one final state and no
   transitions out of this final state. We will denote such NFA by
   NFA'. *)

Inductive nf'_automaton [ A:Set; St:Set; I:(fset A); Q:(fset St); q0:St;
      	       	          f:St; delta:(fset St*(A*St)) ] : Prop :=
  nf'_automaton_intro :
      (member f Q) ->
      (member q0 Q) ->
      ((a:A)(q:St)~(member (f,(a,q)) delta)) ->
      (subset delta (product Q (product I Q))) ->
      (nf'_automaton A St I Q q0 f delta).

(* The set recognized by a NFA'. *)

Inductive nf'_regularset [ A:Set; St:Set; I:(fset A); L:(wordset A) ] : Set :=

  nf'_regular_intro :
      (Q:(fset St))(q0:St)(f:St)
      (delta:(fset St*(A*St)))
      (nf'_automaton I Q q0 f delta) ->
      (eq_wordset L (nf_accepts I Q q0 (singleton f) delta)) ->
      (nf'_regularset A St I L).
