(**************************************************************************)
(*          Nondeterministic Finite Automata with epsilon-moves           *)
(**************************************************************************)

Implicit Arguments On.

Require FiniteSets.

(* Given an alphabet A and a set St,
   a nondeterministic finite automaton with epsilon-moves (NFE)
   is a 5-tuple (I,Q,q0,F,delta), together with an epsilon, where
      
      o) epsilon is the empty input, and is not a member of I

      i) Q is a finite nonempty set of states (over St)
     ii) I is a finite nonempty set of inputs (over A)
    iii) delta is a map from Q x (I U {epsilon}) to 2^Q (power set of Q)
         Is is equivalent to consider delta as a subset of
         Q x (I U {epsilon}) x Q.
     iv) q0 is the initial state (member of Q)
      v) F is the set of final states (subset of Q).  	*)

Inductive nfe_automaton [ A:Set; St:Set; eps:A; I:(fset A); Q:(fset St); q0:St;
      	       	          F:(fset St); delta:(fset St*(A*St)) ] : Prop :=
  nfe_automaton_intro :
     ~(member eps I) ->
      (member q0 Q) ->
      (subset F Q) ->
      (subset delta (product Q (product (add eps I) Q))) ->
      (nfe_automaton A St eps I Q q0 F delta).

(**************************************************************************)
(*                       The set of accepted words                        *)
(**************************************************************************)

Require Words.

(* First we define the notion of path: there is a path in an automaton A
   from the state e1 to the state e2 for the word w iff delta(e1,w)=e2.	*)

Inductive nfe_path [ A:Set; St:Set; eps:A; I:(fset A); Q:(fset St);
      	       	     delta:(fset St*(A*St)) ] : St -> St -> (word A) -> Prop :=

  nfe_path_nil  : (e:St)(member e Q)
               -> (nfe_path A St eps I Q delta e e (nil A))

| nfe_path_cons : (e1,e2,e:St)(w:(word A))(a:A)
       	          (nfe_path A St eps I Q delta e1 e2 w) ->
	          (member a I) ->
	          (member (e,(a,e1)) delta) ->
	          (member e Q) ->
	          (nfe_path A St eps I Q delta e e2 (cons A a w))

| nfe_path_epsilon : (e1,e2,e:St)(w:(word A))
                     (nfe_path A St eps I Q delta e1 e2 w) ->
		     (member (e,(eps,e1)) delta) ->
		     (member e Q) ->
		     (nfe_path A St eps I Q delta e e2 w).

(* A finite automaton A accepts a word w if A, starting in q0, goes through
   some computation and reads all of w, and stops in some final state.	
   
   The set of all accepted words is (nfe_accepts A St I Q q0 F delta). *)

Inductive nfe_accepts [ A:Set; St:Set; eps:A; I:(fset A); Q:(fset St); q0:St;
      	       	        F:(fset St); delta:(fset St*(A*St)); w:(word A)] : Prop
  :=
  nfe_accepts_intro : (f:St) (member f F) ->
      	       	      (nfe_path eps I Q delta q0 f w) ->
		      (nfe_accepts A St eps I Q q0 F delta w).

Require Sets_of_words.

(* A set of words L is called nfe_regular is there is some
   nondeterministic finite automaton with epsilon-moves A
   such that L=T(A).
   We will prove later that it corresponds to the notion of
   nf_regular sets, so to th notion of regular sets. *)

Inductive nfe_regularset [ A:Set; St:Set; eps:A; I:(fset A);
                           L:(wordset A) ] : Set :=

  nfe_regular_intro :
      (Q:(fset St))(q0:St)(F:(fset St))
      (delta:(fset St*(A*St)))
      (nfe_automaton eps I Q q0 F delta) ->
      (eq_wordset L (nfe_accepts eps I Q q0 F delta)) ->
      (nfe_regularset A St eps I L).

(* To do the proof of Kleene's theorem, it is convenient to introduce
   a partiular case of NFA with epsilon-moves : those having one final
   state and no transitions out of this final state.
   We will denote such NFE by NFE'. *)

Inductive nfe'_automaton
                [ A:Set; St:Set; eps:A; I:(fset A); Q:(fset St); q0:St;
      	       	  f:St; delta:(fset St*(A*St)) ] : Prop :=
  nfe'_automaton_intro :
      (member f Q) ->
     ~(member eps I) ->
      (member q0 Q) ->
      ((a:A)(q:St)~(member (f,(a,q)) delta)) ->
      (subset delta (product Q (product (add eps I) Q))) ->
      (nfe'_automaton A St eps I Q q0 f delta).

Require Properties.

(* A NFE' is clearly a NFE. *)

Lemma nfe'_automaton_is_nfe_automaton : 
   (A,St:Set)(eps:A)(I:(fset A))(Q:(fset St))(q0:St)
   (f:St)(delta:(fset St*(A*St)))

   (nfe'_automaton eps I Q q0 f delta) ->
   (nfe_automaton  eps I Q q0 (singleton f) delta).

(* The set recognized by a NFE' *)

Inductive nfe'_regularset [ A:Set; St:Set; eps:A; I:(fset A);
                            L:(wordset A) ] : Set :=

  nfe'_regular_intro :
      (Q:(fset St))(q0:St)(f:St)
      (delta:(fset St*(A*St)))
      (nfe'_automaton eps I Q q0 f delta) ->
      (eq_wordset L (nfe_accepts eps I Q q0 (singleton f) delta)) ->
      (nfe'_regularset A St eps I L).

(* A set recognized by a NFE' is clearly recognized by a NFE,
   since a NFE' is a NFE. *)

Lemma nfe'_regular_is_nfe_regular :
   (A,St:Set)(eps:A)(I:(fset A))(L:(wordset A))
   (nfe'_regularset St eps I L) -> (nfe_regularset St eps I L).

