(**************************************************************************)
(*              The family of regular sets if closed under *              *)
(**************************************************************************)

Implicit Arguments On.

Require NFA_eps.
Require NatProperties.
Require MoreProperties.

Section Star.

Variable A : Set.
Variable I : (fset A).
Variable eps : A.
Hypothesis eps_not_in_I : ~(member eps I).

(* Here we define the NFE' automaton that recognizes l* . *)

Section Aut_for_star.

Variable l1 : (wordset A).

Variable Q1 : (fset nat).
Variable q0_1 : nat.
Variable f1 : nat.
Variable delta1 : (fset nat*A*nat).

Hypothesis aut1 : (nfe'_automaton eps I Q1 q0_1 f1 delta1).

(* First we construct the elements of the automaton
   (Q,q0,f0 and delta). *)

Definition exists_f0 := (outside Q1).

Definition f0 : nat := (proj1_sig ? ? exists_f0).

Definition Q := (add f0 Q1).

Definition q0 := q0_1.

Definition delta :=
  (add (q0,(eps,f0))
  (add (f1,(eps,q0))
  (add (f1,(eps,f0))
  delta1))).

(* Then we prove that it is a NFE' automaton. *)

Lemma member_f0_Q : (member f0 Q).
Proof.
Unfold Q. Auto.
Save.

Lemma member_q0_Q : (member q0 Q).
Proof.
Unfold q0 Q. Apply member_x_add_y.
Elim aut1 ; Auto.
Save.

Lemma f0_not_q0 : ~f0=q0.
Proof.
Generalize (sig_sig exists_f0) .
Change ~(member f0 Q1)->~f0=q0.
Unfold not.
Intros H1 equ.
Apply H1.
Rewrite -> equ.
Unfold q0.
Elim aut1 ; Auto.
Save.
Hint f0_not_q0.

Lemma f0_not_f1 : ~f0=f1.
Proof.
Generalize (sig_sig exists_f0) .
Change ~(member f0 Q1)->~f0=f1.
Unfold not.
Intros H1 equ.
Apply H1.
Rewrite -> equ.
Elim aut1 ; Auto.
Save.
Hint f0_not_f1.

Lemma H_f0 : (a:A)(q:nat)~(member (f0,(a,q)) delta).
Proof.
Intros a q.
Red.
Unfold delta.
Intros Hm. Elim (add_def_1 Hm).

Intros equ.
Injection equ.
Intros.
(Absurd f0=q0; Auto).

Clear Hm. Intro Hm. Elim (add_def_1 Hm).
Intros equ.
Injection equ.
Intros.
(Absurd f0=f1; Auto).

Clear Hm. Intro Hm. Elim (add_def_1 Hm).
Intros equ.
Injection equ.
Intros.
(Absurd f0=f1; Auto).

Clear Hm. Intro Hm.
Absurd (member f0 Q1).
Generalize (sig_sig exists_f0) .
Auto.
Apply prod_left with b:=(product (add eps I) Q1) xb:=(a,q).
Apply member_subset with a:=delta1.
Assumption.
(Elim aut1; Auto).
Save.
 
Lemma H_delta : (subset delta (product Q (product (add eps I) Q))).
Proof.
Unfold subset delta.
Intros x Hm. Elim (add_def_1 Hm).

Clear  Hm.
Intros equ.
Rewrite -> equ.
Apply member_pair_prod.
Exact member_q0_Q. 
Apply member_pair_prod.
Auto.
Unfold Q.
Auto.

Clear  Hm. Intro Hm. Elim (add_def_1 Hm).
Intros equ.
Rewrite -> equ.
Apply member_pair_prod.
Unfold Q.
Apply member_x_add_y.
(Elim aut1; Auto).
Apply member_pair_prod.
Auto.
Unfold Q.
Apply member_x_add_y.
(Elim aut1; Auto).

Clear  Hm. Intro Hm. Elim (add_def_1 Hm).
Intros equ.
Rewrite -> equ.
Apply member_pair_prod.
Unfold Q.
Apply member_x_add_y.
(Elim aut1; Auto).
Apply member_pair_prod.
Auto.
Exact member_f0_Q.

Clear Hm. Intro Hm.
Apply member_subset with a:=delta1.
Assumption.
Unfold Q.
Apply subset_prod33.
Apply subset_prod31.
(Elim aut1; Auto).
Save.

Lemma star_automaton : (nfe'_automaton eps I Q q0 f0 delta).
Proof.
Apply (nfe'_automaton_intro 
      	 member_f0_Q
      	 eps_not_in_I 
      	 member_q0_Q
      	 H_f0
      	 H_delta).
Save.


Axiom accepts_star :
  (w:(word A))
  (wordset_star l1 w) ->
  (nfe_accepts eps I Q q0 (singleton f0) delta w).

Axiom splitting : 
  (w:(word A))
  (nfe_accepts eps I Q q0 (singleton f0) delta w) ->
  (wordset_star l1 w).

(* Then we prove that the recognized language is l1* . *) 

Lemma star_is_reg : 
  (eq_wordset l1 (nfe_accepts eps I Q1 q0_1 (singleton f1) delta1)) ->
  (eq_wordset (wordset_star l1)
              (nfe_accepts eps I Q q0 (singleton f0) delta)).
Proof.
Intros H_l1.
Unfold eq_wordset.
Intros w.
Split.

Intro. Apply accepts_star. Assumption.

Intro. Apply splitting. Assumption.
Save.

End Aut_for_star.


(* We prove here that the family of regular sets is closed under
   the Kleene closure. *)

Theorem closed_under_star :
  (l:(wordset A))
  (regexp I l)
    ->(nfe'_regularset nat eps I l)
      ->(nfe'_regularset nat eps I (wordset_star l)).
Proof.
Intros l1 reg_l1 nfe'_l1.
Elim nfe'_l1. Intros Q1 q0_1 f1 delta1 aut1 H_l1.
Apply nfe'_regular_intro
  with Q  := (Q Q1)
       q0 := (q0 q0_1)
       f  := (f0 Q1)
       delta := (delta Q1 q0_1 f1 delta1).
Apply star_automaton ; Assumption.
Apply star_is_reg ; Assumption.
Save.


End Star.
