(**************************************************************************)
(*            The family of regular sets if closed under union            *)
(**************************************************************************)

Implicit Arguments On.

Require NFA_eps.
Require DisjUnion.
Require NatProperties.

(* We prove here that the family of regular sets is closed under union 	*)

Section Union.

Variable A : Set.
Variable I : (fset A).
Variable eps : A.
Hypothesis eps_not_in_I : ~(member eps I).

(* Here we define the NFE' automaton that recognizes
   the union of l1 and l2. *)

Section Aut_for_union.

Variable l1,l2 : (wordset A).

Variable Q1,Q2 : (fset nat).
Variable q0_1,q0_2 : nat.
Variable f1,f2 : nat.
Variable delta1,delta2 : (fset nat*A*nat).

Hypothesis aut1 : (nfe'_automaton eps I Q1 q0_1 f1 delta1).
Hypothesis aut2 : (nfe'_automaton eps I Q2 q0_2 f2 delta2).

(* First we construct the elements of the automaton
   (Q,q0,f0 and delta). *)

Definition Q1Q2 : (fset nat*bool) := (disj_union Q1 Q2).

Definition exists_q0 := (outside2 Q1Q2).

Definition q0 : nat*bool := (proj1_sig ? ? exists_q0).

Definition exists_f0 := (outside2 (add q0 Q1Q2)).

Definition f0 : nat*bool := (proj1_sig ? ? exists_f0).

Definition Q : (fset nat*bool) := (add q0 (add f0 Q1Q2)).

Definition delta'1 :=
  (map [x:nat*A*nat]Cases x of (q,(a,q')) => (linj q,(a,linj q')) end delta1).

Definition delta'2 :=
  (map [x:nat*A*nat]Cases x of (q,(a,q')) => (rinj q,(a,rinj q')) end delta2).

Definition delta : (fset (nat*bool)*A*(nat*bool)) :=
  (add (q0,(eps,(linj q0_1)))
  (add (q0,(eps,(rinj q0_2)))
  (add (linj f1,(eps,f0))
  (add (rinj f2,(eps,f0))
  (union delta'1 delta'2))))).

(* Then we prove that it is a NFE' automaton. *)

Lemma member_f0_Q : (member f0 Q).

Lemma member_q0_Q : (member q0 Q).

Lemma f0_not_q0 : ~f0=q0.

Lemma f0_not_linj_q1 : (q:nat)(member q Q1) -> ~f0=(linj q).

Lemma f0_not_linj_f1 : ~f0=(linj f1).

Lemma f0_not_rinj_q2 : (q:nat)(member q Q2) -> ~f0=(rinj q).

Lemma f0_not_rinj_f2 : ~f0=(rinj f2).

Lemma f0_notin_delta'1 : (q:nat*bool)(a:A)
  ~(member (f0,(a,q)) delta'1).

Lemma f0_notin_delta'2 : (q:nat*bool)(a:A)
  ~(member (f0,(a,q)) delta'2).

Lemma H_f0 : (a:A)(q:nat*bool)~(member (f0,(a,q)) delta).

Lemma H_delta : (subset delta (product Q (product (add eps I) Q))).

Lemma union_automaton : (nfe'_automaton eps I Q q0 f0 delta).

(* Path extension. 
   If we have a path in automaton A1 or A2, then we have the corresponding
   path in the automaton A. *)

Axiom path_extension_1 :
  (w:(word A))(q,q':nat)
  (nfe_path eps I Q1 delta1 q        q'        w) ->
  (nfe_path eps I Q  delta  (linj q) (linj q') w).

Axiom path_final_extension_1 :
  (w:(word A))
  (nfe_path eps I Q delta (linj q0_1) (linj f1) w) ->
  (nfe_path eps I Q delta q0          f0        w).

Axiom path_extension_2 :
  (w:(word A))(q,q':nat)
  (nfe_path eps I Q2 delta2 q        q'        w) ->
  (nfe_path eps I Q  delta  (rinj q) (rinj q') w).

Axiom path_final_extension_2 :
  (w:(word A))
  (nfe_path eps I Q delta (rinj q0_2) (rinj f2) w) ->
  (nfe_path eps I Q delta q0          f0        w).

(* Path restriction.
   If we have a path in the new automaton from q0 to f0 for w,
   then there is either a path from q0_1 to f1 in A1 for w
   or a path from q0_2 to f2 in A2. *)

Axiom path_restriction :
  (w:(word A))
  (nfe_path eps I Q delta q0 f0 w) ->
  (nfe_path eps I Q1 delta1 q0_1 f1 w) \/ (nfe_path eps I Q2 delta2 q0_2 f2 w).

(* Then we prove that the recognized language is the union of l1 and l2. *)

Lemma union_is_reg : 
  (eq_wordset l1 (nfe_accepts eps I Q1 q0_1 (singleton f1) delta1)) ->
  (eq_wordset l2 (nfe_accepts eps I Q2 q0_2 (singleton f2) delta2)) ->
  (eq_wordset (wordset_union l1 l2)
              (nfe_accepts eps I Q q0 (singleton f0) delta)).

End Aut_for_union.

Theorem closed_under_union :
  (l1,l2:(wordset A))
  (regexp I l1)
   ->(nfe'_regularset nat eps I l1)
      ->(regexp I l2)
         ->(nfe'_regularset nat eps I l2)
      	   ->(nfe'_regularset nat*bool eps I (wordset_union l1 l2)).

End Union.
