(**************************************************************************)
(*                 The set {w}, with w in A*, is regular                  *)
(**************************************************************************)
(*
 * On the paper, the proof is rather trivial:
 * if the word w is a1.a2...an then we take the (NF) automaton
 *
 *    q0  -a1->  q1  -a2->  q2  ...  -an->  qn,		with F={qn}
 *
 * The proof is done by induction over w:
 *
 * - if w=nil  then the automaton is ({q0},q0,{q0})
 * - if w=a.w' then by induction hypothesis we have an automaton A=(Q,q0,F,d)
 *   such that L(A)={w'}. We take a new state q outside Q and construct the
 *   automaton A={ QU{q}, q, F, dU{(q,a,q0)} }.
 **************************************************************************)

Implicit Arguments On.

Require FiniteSets.
Require Properties.
Require MoreProperties.
Require NatProperties.
Require NFA.
Require NFA_eps.

Section one_word.

Variable A : Set.
Variable I : (fset A).

(**************************************************************************)
(*         The set reduced to the null word (epsilon) is regular.         *)
(**************************************************************************)
(*
 * To prove this statement, we take the automaton reduced to a unique
 * state O, that is A = { {O}, O, {O}, {} }.
 *
 **************************************************************************)

Section null_word.

Local Q  := (singleton O).
Local q0 := O.
Local f  := O.
Local delta := (empty_set nat*A*nat).

(* Then we prove that it is a NFA' automaton. (The proofs are all 
   trivial with this very simple automaton). *)

Local member_f_Q : (member f Q).
Proof. Unfold f Q. Auto. Save.

Local member_q0_Q : (member q0 Q).
Proof. Unfold q0. Unfold Q. Auto. Save.

Local H_f : (a:A)(q:nat)~(member (f,(a,q))delta).
Proof. Unfold delta. Auto. Save.

Local H_delta : (subset delta (product Q (product I Q))).
Proof. Unfold delta. Auto. Save.

Local is_automaton : (nf'_automaton I Q q0 f delta).
Proof.
Apply (nf'_automaton_intro member_f_Q member_q0_Q H_f H_delta).
Save.

(* Then we prove that it recognizes exactly the language reduced
   to the word nil. *)

Local eq_wordset : (eq_wordset (one_word_set (nil A))
      	       	       	       (nf_accepts I Q q0 (singleton f) delta)).
Proof.
Unfold eq_wordset.
Intros w.
Unfold one_word_set.
Split.
  Intros eq_nil_w.
  Rewrite <- eq_nil_w.
  EApply nf_accepts_intro with f:=q0.
  Unfold q0 f ; Auto.
  Apply nf_path_nil.
  (Unfold q0 Q; Auto).

  Intros H_acc.
  Elim H_acc.
  Intros f' H_f'.
  Case w.
  Auto.
  Intros a l H_cons.
  Elim H_cons.
  Auto.
  Unfold delta.
  Intros.
  (Absurd (member (e,(a0,e1)) (empty_set nat*A*nat)); Auto).
Save.

Lemma null_word_regular : (nf'_regularset nat I (one_word_set (nil A))).

End null_word.

(**************************************************************************)
(*        If the set {w} is regular then the set {aw} is regular.         *)
(**************************************************************************)
(*
 * See the construction above.
 *
 **************************************************************************)

Section cons_word.

Variable a : A.
Hypothesis member_a_I : (member a I).
Variable w : (word A).
Hypothesis monoid_I_w : (monoid I w).

Hypothesis reg_w : (nf'_regularset nat I (one_word_set w)).

Section cons_word1.

Hypothesis  Q : (fset nat).
Hypothesis  q0 : nat.
Hypothesis  f : nat.
Hypothesis  delta : (fset nat*A*nat).
Hypothesis  H_aut : (nf'_automaton I Q q0 f delta).
Hypothesis  H_eq : (eq_wordset (one_word_set w)
                               (nf_accepts I Q q0 (singleton f) delta)).
Hypothesis  nq : nat.
Hypothesis  H_nq : ~(member nq Q).

(* If we have a path in the automaton A, then we still have this path
 * in the extended automaton. *)

Lemma automaton_extension : (w0:(word A))(q,f':nat)
  (member f' (singleton f)) ->
  (nf_path I Q delta q f' w0) ->
  (nf_path I (add nq Q)
     (add (nq,(a,q0)) delta) q f' w0).

(* The extended automaton accepts the word a.w  *)

Lemma cons_regular1 :
   (nf_accepts I (add nq Q) nq (singleton f)
     (add (nq,(a,q0)) delta) (cons A a w)).

(* Any transition of the form (nq,_,_) in the extended automaton is
 * necessarily the transition (nq,a,q0). *)

Lemma first_transition : (a0:A)(e1:nat)
  (member (nq,(a0,e1)) (add (nq,(a,q0)) delta)) ->
  a0=a /\ e1=q0.

(* If we have a path in the extended automaton starting from of state
 * of Q, then it is also a path in the initial automaton. *)

Lemma path_restricted_to_A : (l:(word A))(q,f':nat)
  (member q Q) ->
  (nf_path I (add nq Q)
           (add (nq,(a,q0)) delta) q f' l) ->
  (nf_path I Q delta q f' l).

(* If a word is accepted by the extended automaton, it is necessarily
 * the word a.w	 *)

Lemma cons_regular2 : (w0:(word A))
  (nf_accepts I (add nq Q) nq (singleton f)
     (add (nq,(a,q0)) delta) w0)->(cons A a w)=w0.

End cons_word1.

(* The set {a.w} is regular. *)

Lemma cons_regular :
  (nf'_regularset nat I (one_word_set (cons A a w))).

End cons_word.

(**************************************************************************)
(*                        And the expected result:                        *)
(**************************************************************************)

Lemma one_word_nf_regular :
  (w:(word A)) (monoid I w) -> (nf'_regularset nat I (one_word_set w)).

Require NFA_is_NFE.

Lemma one_word_regular :
  (eps:A) ~(member eps I) ->
  (w:(word A)) (monoid I w) -> (nfe'_regularset nat eps I (one_word_set w)).

End one_word.

