(**************************************************************************)
(*                 The set {w}, with w in A*, is regular                  *)
(**************************************************************************)
(*
 * On the paper, the proof is rather trivial:
 * if the word w is a1.a2...an then we take the (NF) automaton
 *
 *    q0  -a1->  q1  -a2->  q2  ...  -an->  qn,		with F={qn}
 *
 * The proof is done by induction over w:
 *
 * - if w=nil  then the automaton is ({q0},q0,{q0})
 * - if w=a.w' then by induction hypothesis we have an automaton A=(Q,q0,F,d)
 *   such that L(A)={w'}. We take a new state q outside Q and construct the
 *   automaton A={ QU{q}, q, F, dU{(q,a,q0)} }.
 **************************************************************************)

Implicit Arguments On.

Require FiniteSets.
Require Properties.
Require MoreProperties.
Require NatProperties.
Require NFA.
Require NFA_eps.


Section one_word.

Variable A : Set.
Variable I : (fset A).


(**************************************************************************)
(*         The set reduced to the null word (epsilon) is regular.         *)
(**************************************************************************)
(*
 * To prove this statement, we take the automaton reduced to a unique
 * state O, that is A = { {O}, O, {O}, {} }.
 *
 **************************************************************************)

Section null_word.

Local Q  := (singleton O).
Local q0 := O.
Local f  := O.
Local delta := (empty_set nat*A*nat).

(* Then we prove that it is a NFA' automaton. (The proofs are all 
   trivial with this very simple automaton). *)

Local member_f_Q : (member f Q).
Proof. Unfold f Q. Auto. Save.

Local member_q0_Q : (member q0 Q).
Proof. Unfold q0. Unfold Q. Auto. Save.

Local H_f : (a:A)(q:nat)~(member (f,(a,q))delta).
Proof. Unfold delta. Auto. Save.

Local H_delta : (subset delta (product Q (product I Q))).
Proof. Unfold delta. Auto. Save.

Local is_automaton : (nf'_automaton I Q q0 f delta).
Proof.
Apply (nf'_automaton_intro member_f_Q member_q0_Q H_f H_delta). 
Save.


(* Then we prove that it recognizes exactly the language reduced
   to the word nil. *)

Local eq_wordset : (eq_wordset (one_word_set (nil A))
      	       	       	       (nf_accepts I Q q0 (singleton f) delta)).
Proof.
Unfold eq_wordset.
Intros w.
Unfold one_word_set.
Split.
  Intros eq_nil_w.
  Rewrite <- eq_nil_w.
  EApply nf_accepts_intro with f:=q0.
  Unfold q0 f ; Auto.
  Apply nf_path_nil.
  (Unfold q0 Q; Auto).

  Intros H_acc.
  Elim H_acc.
  Intros f' H_f'.
  Case w.
  Auto.
  Intros a l H_cons.
  Elim H_cons.
  Auto.
  Unfold delta.
  Intros.
  (Absurd (member (e,(a0,e1)) (empty_set nat*A*nat)); Auto).
Save.


Lemma null_word_regular : (nf'_regularset nat I (one_word_set (nil A))).
Proof.
Apply (nf'_regular_intro is_automaton eq_wordset).         
Save.

End null_word.


(**************************************************************************)
(*        If the set {w} is regular then the set {aw} is regular.         *)
(**************************************************************************)
(*
 * See the construction above.
 *
 **************************************************************************)

Section cons_word.

Variable a : A.
Hypothesis member_a_I : (member a I).
Variable w : (word A).
Hypothesis monoid_I_w : (monoid I w).

Hypothesis reg_w : (nf'_regularset nat I (one_word_set w)).


Section cons_word1.

Hypothesis  Q : (fset nat).
Hypothesis  q0 : nat.
Hypothesis  f : nat.
Hypothesis  delta : (fset nat*A*nat).
Hypothesis  H_aut : (nf'_automaton I Q q0 f delta).
Hypothesis  H_eq : (eq_wordset (one_word_set w)
                               (nf_accepts I Q q0 (singleton f) delta)).
Hypothesis  nq : nat.
Hypothesis  H_nq : ~(member nq Q).


(* If we have a path in the automaton A, then we still have this path
 * in the extended automaton. *)

Lemma automaton_extension : (w0:(word A))(q,f':nat)
  (member f' (singleton f)) ->
  (nf_path I Q delta q f' w0) ->
  (nf_path I (add nq Q)
     (add (nq,(a,q0)) delta) q f' w0).
Proof.
Intros w0 q f' H_f' H_path.
Elim H_path.
Intros.
(Apply nf_path_nil; Auto).

Intros.
(Apply nf_path_cons with e1; Auto).
Save.

(* The extended automaton accepts the word a.w  *)

Lemma cons_regular1 :
   (nf_accepts I (add nq Q) nq (singleton f)
     (add (nq,(a,q0)) delta) (cons A a w)).
Proof.
Unfold eq_wordset in H_eq.
Elim (H_eq w).
Unfold one_word_set. Intros H1 H2.
Elim (H1 (refl_equal ? w)).
Intros f' H_f'.
Intro H_path.
Apply nf_accepts_intro with f:=f'.
Assumption.
 
(Apply nf_path_cons with q0; Auto).
Apply automaton_extension; Assumption.
Save.

(* Any transition of the form (nq,_,_) in the extended automaton is
 * necessarily the transition (nq,a,q0). *)

Lemma first_transition : (a0:A)(e1:nat)
  (member (nq,(a0,e1)) (add (nq,(a,q0)) delta)) ->
  a0=a /\ e1=q0.
Proof.
Intros a0 e1 H_bel.
Elim (add_def_1 H_bel).
Intros eq.
Injection eq.
Auto.
 
Intros H_bel2.
(Absurd (member nq Q); Auto).
Apply prod_left with b:=(product I Q) xb:=(a0,e1).
Apply member_subset with a:=delta.
Assumption.
Elim H_aut ; Auto.
Save.

(* If we have a path in the extended automaton starting from of state
 * of Q, then it is also a path in the initial automaton. *)

Lemma path_restricted_to_A : (l:(word A))(q,f':nat)
  (member q Q) ->
  (nf_path I (add nq Q)
           (add (nq,(a,q0)) delta) q f' l) ->
  (nf_path I Q delta q f' l).
Proof.
Induction l.

(* 1. l is nil *)
Intros q f' H_q H_p.
Inversion H_p.

  (* 1.1 *)
  Apply nf_path_nil.
  Rewrite <- H1.
  Assumption.
  (* 1.2 *)
  Absurd (cons A a0 w0)=(nil A).
  Apply sym_not_equal.
  Apply nil_cons.
  Assumption.

(* 2. l is cons *)
Intros a0 l0 H_l0 q f' H_q H_f'.
Inversion H_f'.
(Absurd (nil A)=(cons A a0 l0); Auto).
Apply nil_cons.
 
Cut (member (q,(a0,e1)) delta).
Intro H_bel.
Cut (member q Q)/\(member (a0,e1) (product I Q)).
Intros conj. Elim conj. Clear  conj. Intros.
Apply nf_path_cons with e1 ; Auto.
(Apply (H_l0 e1 f'); Auto).
Apply prod_right with a:=I xa:=a0. Assumption.
Split. Assumption.
Apply member_pair_prod.
Assumption.
Apply prod_right with a:=I xa:=a0.
Apply prod_right with a:=Q xa:=q.
Apply member_subset with a:=delta.
Assumption.
(Elim H_aut; Auto).
Elim (add_def_1 H5).
Intro equ. Injection equ.
Intros. Absurd (member nq Q) ; Auto.
Rewrite <- H9. Assumption.
Auto.
Save.

(* If a word is accepted by the extended automaton, it is necessarily
 * the word a.w	 *)

Lemma cons_regular2 : (w0:(word A))
  (nf_accepts I (add nq Q) nq (singleton f)
     (add (nq,(a,q0)) delta) w0)->(cons A a w)=w0.
Proof.
Intro w0. Case w0. 

Intro H_acc. Inversion H_acc.
Inversion H0.
(Absurd (member nq Q); Auto).
Rewrite H3.
Apply member_subset with a:=(singleton f).
Assumption.
Elim H_aut. Intros.
Apply subset_singleton with f:=f ; Auto.
Absurd (cons A a0 w1)=(nil A) ; Auto.
Apply sym_not_equal. Apply nil_cons.  (* bug of Discriminate *)

Intros a0 l H_acc. Inversion H_acc.
Inversion H0.
Absurd (nil A)=(cons A a0 l) ; Auto.
Apply nil_cons.
Elim first_transition with a0:=a0 e1:=e1.
2:Assumption.
Intros eq_a0_a eq_e1_q0.
Rewrite eq_a0_a.
Apply f_equal with f:=[x:(word A)](cons A a x).
Unfold eq_wordset in H_eq.
Elim (H_eq l).
Unfold one_word_set.
Intros Heq1 Heq2.
Apply Heq2.
Apply nf_accepts_intro with f:=f.
Auto.

Rewrite eq_e1_q0 in H3.
Apply path_restricted_to_A.
Elim H_aut ; Auto.
Generalize (member_of_singleton H).
Intro equ. Rewrite <- equ. Assumption.
Save.

End cons_word1.

(* The set {a.w} is regular. *)

Lemma cons_regular :
  (nf'_regularset nat I (one_word_set (cons A a w))).
Proof.
Elim reg_w.
Intros Q q0 f delta H_aut H_eq.
Elim (outside Q). Intros nq H_nq.
Apply nf'_regular_intro 
      with I:=I L:=(one_word_set (cons A a w))
           Q:=(add nq Q) q0:=nq f:=f
           delta:=(add (nq,(a,q0)) delta).
Elim H_aut. Intros. Apply nf'_automaton_intro with f:=f ; Auto.
Intros a0 q. Red.
Intro Hm. Elim (add_def_1 Hm).
Intro equ. Injection equ.
Intros equ1 equ2 equ3. Absurd (member nq Q) ; Auto.
Rewrite <- equ3. Assumption.
Change ~(member (f,(a0,q)) delta). Auto.
Apply add_in_product ; Auto.

Unfold eq_wordset. Unfold one_word_set.
Intro w0. Split.

(* 1. w0=a.w => (accepts A' w0) *)
Intro eq_w0. Rewrite <- eq_w0.
Apply cons_regular1.
Assumption.

(* 2. (accepts A' w0) => w0=a.w *)
Intro H_acc.
Apply (cons_regular2 H_aut H_eq) with nq:=nq ; Assumption.
Save.

End cons_word.


(**************************************************************************)
(*                        And the expected result:                        *)
(**************************************************************************)

Lemma one_word_nf_regular :
  (w:(word A)) (monoid I w) -> (nf'_regularset nat I (one_word_set w)).
Proof.
Induction w.

  Intro. Apply null_word_regular.

  Intros a l H_l H_mon.
  Elim (monoid_inv_cons H_mon). Intros.
  Apply cons_regular.
  Assumption.
  Exact (H_l H0).
Save.


Require NFA_is_NFE.

Lemma one_word_regular :
  (eps:A) ~(member eps I) ->
  (w:(word A)) (monoid I w) -> (nfe'_regularset nat eps I (one_word_set w)).
Proof.
Intros eps H_eps w H_mon.
Apply nfa'_regular_is_nfe'_regular.
Assumption.
Exact (one_word_nf_regular H_mon).
Save.

End one_word.

