(**************************************************************************)
(*                         Finites sets as lists                          *)
(**************************************************************************)

Implicit Arguments On.

Require PolyList.

(* (fset A) : type of finite sets over type A.
   It is the type of polymorphic lists *)

Definition fset : Set -> Set
  := [A:Set](list A).

(* the "belongs to" relation : "In", defined in PolyList *)

Definition member : (A:Set) A -> (fset A) -> Prop
  := In.

(* "belongs" is decidable, if equality is *)

Definition eq_dec : Set -> Set
  := [A:Set] (x,y:A) {x=y} + {~(x=y)}.

Lemma In_dec : (A:Set)
  (eq_dec A) -> (x:A)(a:(fset A)) {(member x a)}+{~(member x a)}.

(* the empty set : nil *)

Definition empty_set : (A:Set) (fset A)
  := nil.

Lemma empty_set_def : (A:Set)(a:A) ~(member a (empty_set A)).

(* the adjunction of one element to a set : cons *)

Definition add : (A:Set) A -> (fset A) -> (fset A)
  := cons.

Lemma add_def : (A:Set)(s:(fset A))(x,y:A)
     (member x (add y s))
 <-> (x=y \/ (member x s)).

(* a set reduced to one element a (the singleton {x})
   is defined in terms of the empty set and add.
   It is the list reduced to the element x. *)

Definition singleton : (A:Set) A -> (fset A) :=
  [A:Set][a:A] (add a (empty_set A)).

(* the union of two sets : concatenation of lists *)

Definition union : (A:Set) (fset A) -> (fset A) -> (fset A)
  := app.

Lemma union_def : (A:Set) (a,a':(fset A)) (x:A)
     (member x (union a a'))
 <-> ((member x a) \/ (member x a')).

(* the intersection of two sets *)

Parameter intersection : (A:Set) (fset A) -> (fset A) -> (fset A).

Axiom intersection_def : (A:Set) (a,a':(fset A)) (x:A)
     (member x (intersection a a'))
 <-> ((member x a) /\ (member x a')).

(* the cartesian product of two sets *)

Fixpoint one_product [ A,B:Set ; a:A ;  l:(fset B) ] : (fset A*B)
  := Cases l of
       nil         => (nil A*B)
     | (cons b l') => (cons ? (a,b) (one_product a l')) end.

Fixpoint product [ A,B:Set ; la:(fset A) ] : (fset B) -> (fset A*B)
  := [lb:(fset B)]Cases la of
       nil          => (nil A*B)
     | (cons a la') => (union (one_product a lb) (product la' lb)) end.

Axiom product_def : (A,B:Set) (a:(fset A)) (b:(fset B)) (x:A*B)
     (member x (product a b))
 <-> (Cases x of (pair xa xb) => (member xa a) /\ (member xb b) end).

(* the sets inclusion *)

Definition subset : (A:Set) (fset A) -> (fset A) -> Prop :=
  [A:Set][a,b:(fset A)] (x:A) (member x a) -> (member x b).

(* the power set : P(s) = the sets of parts of s *)

Parameter powerset : (A:Set) (fset A) -> (fset (fset A)).

Axiom powerset_def : (A:Set) (s:(fset A)) (x:(fset A))
     (member x (powerset s))
 <-> (subset x s).

(* the equality of finite sets, defined as mutual inclusion *)

Definition eq_fset : (A:Set) (fset A) -> (fset A) -> Prop :=
  [A:Set][a,b:(fset A)] (subset a b) /\ (subset b a).

(* Mapping from a finite set a to a finite set b *)

Definition mapping : (A,B:Set) (fset A) -> (fset B) -> Set
  := [A,B:Set] [a:(fset A)][b:(fset B)]
     (x:A)(member x a) -> { y:B | (member y b) }.

Definition map_app : (B:Set)(b:(fset B)) {y:B | (member y b)} -> B
  := [B:Set][b:(fset B)][p:{y:B | (member y b)}] (proj1_sig ? ? p).

(* The set of { f(x) | x in a} for a finite set a. *)

Fixpoint map [ A,B:Set; f:A->B; la:(fset A) ] : (fset B) :=
  Cases la of
    nil          => (nil B)
  | (cons a la') => (cons ? (f a) (map f la')) end.

Axiom map_def : (A,B:Set)(f:A->B)(a:(fset A))
  (y:B) (member y (map f a)) <-> (Ex [x:A]((member x a) /\ y=(f x))).

(* The list of all the elements of a finite sets.
   It is the set itself: list_of_fset is the identity ! *)

Definition list_of_fset : (A:Set)(fset A) -> (list A)
  := [A:Set][a:(fset A)]a.

Lemma list_of_fset_def : (A:Set)(a:(fset A))
  (x:A)(member x a) <-> (In A x (list_of_fset a)).

(* Decomposition principle: a finite set is either the empty set or
   the adjunction of an element to a finite set.
   It is just the fact that lists are only constructible from nil and
   cons constructors. *)

Lemma fset_principle : (A:Set) (a:(fset A))
  (a = (empty_set A)) \/ (Ex [x:A](Ex [a':(fset A)] a = (add x a'))).

(**************************************************************************)
(*             The particular case of finite sets of naturals             *)
(**************************************************************************)

(* In the particular case of natrural numbers, we have decidability
   of the equality (eq_nat_dec, in module Peano_dec). 
   Thus, we can write efficient versions of add (add_nat) and
   union (union_nat) that removes redundant elements.
   We have decidability of member. *)

Require Peano_dec.

Definition In_nat_dec := (In_dec eq_nat_dec).

Definition add_nat : nat -> (fset nat) -> (fset nat)
  := [n:nat][a:(fset nat)]
     Cases (In_nat_dec n a) of
       (left _)  => a
     | (right _) => (cons ? n a) end.

Fixpoint union_nat [ a:(fset nat) ] : (fset nat) -> (fset nat)
  := [b:(fset nat)]
     Cases a of
        nil          => b
     | (cons x a')   => (add_nat x (union_nat a' b))
     end.

