(**************************************************************************)
(*                      Finites Sets (Axiomatization)                     *)
(**************************************************************************)

Implicit Arguments On.


(* (fset A) : type of finite sets over type A *)

Parameter fset : Set -> Set.


(* the "member" relation *)

Parameter member : (A:Set) A -> (fset A) -> Prop.


(* the empty set *)

Parameter empty_set : (A:Set) (fset A).

Axiom empty_set_def : (A:Set)(a:A) ~(member a (empty_set A)).
Hint empty_set_def.


(* the adjunction of one element to a set *)

Parameter add : (A:Set) A -> (fset A) -> (fset A).

Axiom add_def : (A:Set)(s:(fset A))(x,y:A)
     (member x (add y s))
 <-> (x=y \/ (member x s)).


(* a set reduced to one element a (the singleton {x})
   is defined in terms of the empty set and add *)

Definition singleton : (A:Set) A -> (fset A) :=
  [A:Set][a:A] (add a (empty_set A)).


(* the union of two sets *)

Parameter union : (A:Set) (fset A) -> (fset A) -> (fset A).

Axiom union_def : (A:Set) (a,a':(fset A)) (x:A)
     (member x (union a a'))
 <-> ((member x a) \/ (member x a')).


(* the intersection of two sets *)

Parameter intersection : (A:Set) (fset A) -> (fset A) -> (fset A).

Axiom intersection_def : (A:Set) (a,a':(fset A)) (x:A)
     (member x (intersection a a'))
 <-> ((member x a) /\ (member x a')).


(* the cartesian product of two sets *)

Parameter product : (A,B:Set) (fset A) -> (fset B) -> (fset A*B).

Axiom product_def : (A,B:Set) (a:(fset A)) (b:(fset B)) (x:A*B)
     (member x (product a b))
 <-> (Cases x of (pair xa xb) => (member xa a) /\ (member xb b) end).


(* the sets inclusion *)

Definition subset : (A:Set) (fset A) -> (fset A) -> Prop :=
  [A:Set][a,b:(fset A)] (x:A) (member x a) -> (member x b).


(* the power set : P(s) = the sets of parts of s *)

Parameter powerset : (A:Set) (fset A) -> (fset (fset A)).

Axiom powerset_def : (A:Set) (s:(fset A)) (x:(fset A))
     (member x (powerset s))
 <-> (subset x s).


(* the equality of finite sets, defined as mutual inclusion *)

Definition eq_fset : (A:Set) (fset A) -> (fset A) -> Prop :=
  [A:Set][a,b:(fset A)] (subset a b) /\ (subset b a).


(* Mapping from a finite set a to a finite set b *)

Definition mapping : (A,B:Set) (fset A) -> (fset B) -> Set
  := [A,B:Set] [a:(fset A)][b:(fset B)]
     (x:A)(member x a) -> { y:B | (member y b) }.

Definition map_app : (B:Set)(b:(fset B)) {y:B | (member y b)} -> B
  := [B:Set][b:(fset B)][p:{y:B | (member y b)}] (proj1_sig ? ? p).


(* The set of { f(x) | x in a} for a finite set a. *)

Parameter map : (A,B:Set) (A->B) -> (fset A) -> (fset B).

Axiom map_def : (A,B:Set)(f:A->B)(a:(fset A))
  (y:B) (member y (map f a)) <-> (Ex [x:A]((member x a) /\ y=(f x))).


(* The list of all the elements of a finite sets. *)

Require PolyList.

Parameter list_of_fset : (A:Set)(fset A) -> (list A).

Axiom list_of_fset_def : (A:Set)(a:(fset A))
  (x:A)(member x a) <-> (In A x (list_of_fset a)).


(* Decomposition principle: a finite set is either the empty set or
   the adjunction of an element to a finite set. *)

Axiom fset_principle : (A:Set) (a:(fset A))
  (a = (empty_set A)) \/ (Ex [x:A](Ex [a':(fset A)] a = (add x a'))).
