(**************************************************************************)
(*                     Various lemmas on finite sets                      *)
(**************************************************************************)

Implicit Arguments On.

Require FiniteSets.


Section Properties_of_add.

Variable A : Set.
Variable a : (fset A).


Lemma add_def_1 : (x,y:A) (member x (add y a)) -> (x=y \/ (member x a)).
Intros x y. Elim (add_def a x y). Auto.
Save.


Lemma add_def_2 : (x,y:A) (x=y \/ (member x a)) -> (member x (add y a)).
Intros x y. Elim (add_def a x y). Auto.
Save.


Lemma member_x_add_x : (x:A) (member x (add x a)).
Intros x. Apply add_def_2. Auto.
Save.


Lemma member_x_add_y : (x,y:A) (member x a) -> (member x (add y a)).
Intros x y H. Apply add_def_2. Auto.
Save.


Lemma member_add_inv_1 : (x,y:A)
  (member x (add y a)) -> ~x=y -> (member x a).
Intros x y H1 H2. Elim (add_def_1 H1).
Intros H3. Absurd x=y; Auto.
Auto.
Save.

End Properties_of_add.

Hint member_x_add_x.
Hint member_x_add_y.


(* a member to the singleton {a} *)

Lemma member_x_singleton_x : (A:Set) (a:A) (member a (singleton a)).
Unfold singleton. Auto.
Save.

Hint member_x_singleton_x.

Lemma member_of_singleton : (A:Set) (a,x:A)
  (member x (singleton a)) -> x=a.
Proof.
Unfold singleton.
Intros A a x H.
Elim (add_def_1 H).
Auto.
Intro. (Absurd (member x (empty_set A)); Auto).
Save.

Hint member_of_singleton.


Section Properties_of_product.

Variable A,B : Set.
Variable a   : (fset A).
Variable b   : (fset B).


Lemma prod_def_1 : (x:A*B)
    (Cases x of (pair xa xb) => (member xa a)/\(member xb b) end)
 -> (member x (product a b)).
Intro x. Elim (product_def a b x). Auto.
Save.


Lemma prod_def_2 : (x:A*B)
    (member x (product a b))
 -> (Cases x of (pair xa xb) => (member xa a)/\(member xb b) end).
Intro x. Elim (product_def a b x). Auto.
Save.


Lemma member_pair_prod : (xa:A)(xb:B)
  (member xa a) -> (member xb b) ->
  (member (xa,xb) (product a b)).
Intros. Apply prod_def_1. Auto.
Save.


Lemma prod_left : (xa:A)(xb:B)
  (member (xa,xb) (product a b)) -> (member xa a).
Intros xa xb H. Elim (prod_def_2 H). Auto.
Save.


Lemma prod_right : (xa:A)(xb:B)
  (member (xa,xb) (product a b)) -> (member xb b).
Intros xa xb H. Elim (prod_def_2 H). Auto.
Save.


Lemma member_pair_prod_inv : (xa:A)(xb:B)
  (member (xa,xb) (product a b)) ->
  (member xa a) /\ (member xb b).
Intros xa xb H.
Split ; [ Apply (prod_left H) | Apply (prod_right H) ].
Save.

End Properties_of_product.

Hint member_pair_prod.
Hint member_pair_prod_inv.



Section Properties_of_subset.

Variable A,B : Set.
Variable a   : (fset A).
Variable b   : (fset B).


Lemma subset_a_a : (subset a a).
Unfold subset. Auto.
Save.


Lemma subset_a_addxa : (x:A) (subset a (add x a)).
Unfold subset. Auto.
Save.


Lemma subset_trans_add : (a':(fset A)) (x:A)
  (subset a a') -> (subset a (add x a')).
Unfold subset. Auto.
Save.


Lemma subset_add_trans : (a':(fset A)) (x:A)
  (member x a') -> (subset a a') -> (subset (add x a) a').
Unfold subset. Intros.
Elim (add_def_1 H1).
Intros eq. Rewrite -> eq. Auto.
Auto.
Save.


Lemma subset_empty_set : (subset (empty_set A) a).
Unfold subset.
Intros x H.
(Absurd (member x (empty_set A)); Auto).
Save.


Lemma subset_trans_product_left : (d:(fset A*B)) (x:A)
  (subset d (product a b)) ->
  (subset d (product (add x a) b)).
Unfold subset.
Intros d x H x0 H1.
Cut (member x0 (product a b)).
Elim x0. Intros.
Apply member_pair_prod.
Apply member_x_add_y.
Apply (prod_left H0).
Apply (prod_right H0).
Auto.
Save.


Lemma subset_trans_product_right : (d:(fset A*B)) (x:B)
  (subset d (product a b)) ->
  (subset d (product a (add x b))).
Unfold subset.
Intros d x H x0 H1.
Cut (member x0 (product a b)).
Elim x0. Intros.
Apply member_pair_prod.
Apply (prod_left H0).
Apply member_x_add_y.
Apply (prod_right H0).
Auto.
Save.


Lemma member_subset : (a':(fset A)) (x:A)
  (member x a) -> (subset a a') -> (member x a').
Unfold subset. Auto.
Save.
 
End Properties_of_subset.

Hint subset_a_a.
Hint subset_a_addxa.
Hint subset_trans_add.
Hint subset_add_trans.
Hint subset_empty_set.
Hint subset_trans_product_left.
Hint subset_trans_product_right.


Section Properties_of_eq_fset.

Variable A' : Set.
Variable a : (fset A').

Lemma eq_fset_refl : (eq_fset a a).
Proof.
Unfold eq_fset. Auto.
Save.

End Properties_of_eq_fset.

Hint eq_fset_refl.


Lemma subset_singleton2 :
  (A:Set)(f:A)(F,Q:(fset A))
  (eq_fset F (singleton f)) -> 
  (member f Q) ->
  (subset F Q).
Proof.
Intros A f F Q.
Unfold eq_fset subset.
Intros H.
Elim H.
Clear  H.
Intros HF1 HF2.
Intros H_f x H_x.
Generalize (HF1 x H_x) .
Unfold singleton.
Intros Hxadd.
Elim (add_def_1 Hxadd).
Intros equ.
Rewrite -> equ.
Assumption.
Intros He.
(Absurd (member x (empty_set A)); Auto).
Save.


Lemma subset_singleton :
  (A:Set)(f:A)(Q:(fset A))
  (member f Q) ->
  (subset (singleton f) Q).
Proof.
Intros A f F Q.
Unfold subset.
Intros x H.
Elim (member_of_singleton H).
Generalize (member_of_singleton H) .
Intros equ.
Rewrite -> equ.
Assumption.
Save.


(* Properties of union *)

Lemma union_def_1 : (A:Set) (a,a':(fset A)) (x:A)
     (member x (union a a'))
  -> ((member x a) \/ (member x a')).
Proof.
Intros A a a' x Hx.
Elim (union_def a a' x).
Auto.
Save.

Lemma union_def_2 : (A:Set) (a,a':(fset A)) (x:A)
    ((member x a) \/ (member x a'))
 -> (member x (union a a')).
Proof.
Intros A a a' x Hx.
Elim (union_def a a' x).
Auto.
Save.


(* Properties of map *)

Lemma map_def_1 : (A,B:Set)(f:A->B)(a:(fset A))
  (x:A) (member x a) -> (member (f x) (map f a)).
Proof.
Intros A B f a x H.
Elim (map_def f a (f x)).
Intros H1 H2. Apply H2.
Exists x. Auto.
Save.

Lemma map_def_2 : (A,B:Set)(f:A->B)(a:(fset A))
  (y:B) (member y (map f a)) -> (Ex [x:A]((member x a) /\ y=(f x))).
Proof.
Intros A B f a y H.
Elim (map_def f a y).
Auto.
Save.

Lemma map_def_3 : (A,B:Set)(f:A->B)(a:(fset A))
  (y:B) (Ex [x:A]((member x a) /\ y=(f x))) -> (member y (map f a)).
Proof.
Intros A B f a y H.
Elim (map_def f a y).
Auto.
Save.


(* Properties of list_of_fset *)

Lemma list_of_def_1 : (A:Set)(a:(fset A))
  (x:A) (member x a) -> (In A x (list_of_fset a)).
Proof.
Intros A a x.
Elim (list_of_fset_def a x).
Auto.
Save.

Lemma list_of_def_2 : (A:Set)(a:(fset A))
  (x:A) (In A x (list_of_fset a)) -> (member x a).
Proof.
Intros A a x.
Elim (list_of_fset_def a x).
Auto.
Save.
