(**************************************************************************)
(*                             Sets of words                              *)
(**************************************************************************)

Implicit Arguments On.

Require Words.


(* We represent sets of words (over a type A) as their
   characteristic function, that is: *)

Definition wordset := [A:Set] (word A) -> Prop.


(* Equality over sets of words, and its properties. *)

Section Equality_over_sets_of_words.

Variable A : Set. 	(* given an alphabet A *)

Definition eq_wordset : (wordset A) -> (wordset A) -> Prop := 
  [l1,l2 : (wordset A)]
    (w:(word A)) (   ((l1 w)->(l2 w))
                  /\ ((l2 w)->(l1 w))).

Lemma eq_wordset_refl : (L:(wordset A)) (eq_wordset L L).
Red. Auto.
Save.

Lemma eq_wordset_sym : (l1,l2:(wordset A))
  (eq_wordset l1 l2) -> (eq_wordset l2 l1).
Unfold eq_wordset.
Intros l1 l2 H w.
Elim (H w). Auto.
Save.

Lemma eq_wordset_trans : (l1,l2,l3:(wordset A))
  (eq_wordset l1 l2) -> (eq_wordset l2 l3) -> (eq_wordset l1 l3).
Unfold eq_wordset.
Intros l1 l2 l3 H H0 w.
Elim (H0 w).
Elim (H w).
Auto.
Save.

End Equality_over_sets_of_words.


(* The empty set of words *)

Definition empty_wordset : (A:Set) (wordset A)
  := [A:Set][w:(word A)]False.


(* Sets of words reduced to a single word. *)

Definition one_word_set : (A:Set) (word A) -> (wordset A) :=
  [A:Set] [w,w1:(word A)] w=w1.


(* Basic operations on sets of words : union, concatenation and *. *)

Section Operations.

Variable A : Set.


(* concatenation *)

Inductive wordset_concat [l1,l2 : (wordset A)] : (wordset A) :=
  wordset_concat_intro : (w1,w2:(word A)) (l1 w1) -> (l2 w2) 
      	       	      -> (wordset_concat l1 l2 (app A w1 w2)).


(* union *)

Inductive wordset_union [l1,l2 : (wordset A)] : (wordset A) := 
  wordset_union_g : (w:(word A)) (l1 w) -> (wordset_union l1 l2 w)
| wordset_union_d : (w:(word A)) (l2 w) -> (wordset_union l1 l2 w).


(* Kleene closure.
   
   First we define L^n as: 1)  L^O     = {epsilon}
      	       	       	   2)  L^(S n) = L^n . L  *)

Inductive wordset_power [ l : (wordset A) ] : nat -> (wordset A) :=
  wordset_power_O : (w:(word A)) (one_word_set (nil A) w)
      	       	 -> (wordset_power l O w)
| wordset_power_S : (n:nat)(w:(word A))
      	       	    (wordset_concat l (wordset_power l n) w)
      	       	 -> (wordset_power l (S n) w).

(* then L* as the union over the natural numbers of L^n. *)

Inductive wordset_star [ l : (wordset A) ] : (wordset A) :=
  wordset_star_intro : (n:nat)(w:(word A)) (wordset_power l n w)
      	       	    -> (wordset_star l w).

End Operations.


(* There is an induction principle associated to L* :

   If a property P on words holds for every word of L^n then it holds
   for every word of L*.  *)

Section Induction_star.

Variable A : Set.
Variable P : (word A) -> Prop.
Variable l : (wordset A).

Lemma induction_star :
     ( (n:nat)(w:(word A)) (wordset_power l n w)->(P w) ) 
  -> ( (w:(word A))(wordset_star l w)->(P w) ).
Intros. Elim H0 ; Auto.
Save.

End Induction_star.



Require FiniteSets.
Require Properties.

(* The monoid A* for a finite alphabet A (over a type T) *)

Inductive monoid [T:Set; A:(fset T)] : (wordset T) :=

  monoid_nil  : (monoid T A (nil T))

| monoid_cons : (a:T)(w:(word T))
      	       	  (monoid T A w) ->
		  (member a A) ->
		  (monoid T A (cons ? a w)).


Lemma monoid_inv_cons : (T:Set)(A:(fset T))(a:T)(w:(word T))
  (monoid A (cons T a w)) -> (member a A) /\ (monoid A w).
Proof.
Intros T A a w H.
Inversion H.
Auto.
Save.

Lemma monoid_inv_cons_w : (T:Set)(A:(fset T))(a:T)(w:(word T))
  (monoid A (cons T a w)) -> (monoid A w).
Proof.
Intros T A a w H.
Inversion H.
Auto.
Save.

Lemma monoid_inv_cons_a : (T:Set)(A:(fset T))(a:T)(w:(word T))
  (monoid A (cons T a w)) -> (member a A).
Proof.
Intros T A a w H.
Inversion H.
Auto.
Save.

(* The property for a set of words L to belong to the monoid A*,
   A being of finite alphabet (over a type T). *)

Definition islanguage : (T:Set) (fset T) -> (wordset T) -> Prop :=
  [T:Set][A:(fset T)][L:(wordset T)]
    (w:(word T)) (L w) -> (monoid A w).


(* Regular expressions.

   Let I be an alphabet (a finite set, over type T).
   The set of regular expressions over the alphabet I is defined
   as the smallest family:
     - containing the empty set,
     - containing the set {w} for every word w on I
     - which is closed under union, concatenation and *. 

   We define the type regexp of regular expressions together with the
   languages they define. The type regexp is defined in sort Set because
   it has an informative meaning. *)

Inductive regexp [T:Set; I:(fset T)] : (wordset T) -> Set :=

    reg_empty : (regexp T I [w:(word T)]False )
 
  | reg_word  : (w :(word T))
      	       	  (monoid I w) ->
                  (regexp T I (one_word_set w))

  | reg_union : (l1,l2 : (wordset T))
                  (regexp T I l1) -> (regexp T I l2) -> 
                  (regexp T I (wordset_union l1 l2))

  | reg_conc  : (l1,l2 : (wordset T))
                  (regexp T I l1) -> (regexp T I l2) -> 
                  (regexp T I (wordset_concat l1 l2))

  | reg_star  : (l : (wordset T))
                  (regexp T I l) -> (regexp T I (wordset_star l)).


(* The first projection of (sig A P) satisfies P.
   Must belong to the standard library. *)

Lemma sig_sig : (A:Set)(P:A->Prop)(H:(sig A P))
  (P (proj1_sig A P H)).
Proof.
Intros A P H. Elim H. Auto.
Save.


(* Given a transition function delta : Q x I -> Q, we extend it to
   apply to a state and a string rather than a state and a symbol, with

      extw delta nil        q = q
      extw delta (cons a w) q = extw delta w (delta (a,q))

   *)

Section extw_def.

Variable A,St : Set.
Variable Q : (fset St).
Variable I : (fset A).
Variable delta : (mapping (product Q I) Q).


Lemma extw : (w:(word A))(monoid I w) -> (mapping Q Q).
Proof.
Induction w.
Intros Hw.
Unfold mapping.
Intros q Hq.
Exists  q.
Exact Hq.

Intros a w' extw_w' H_cons.
Unfold mapping.
Intros q Hq.
Cut (member (q,a) (product Q I)).
Intro Hqa.
Apply (extw_w' (monoid_inv_cons_w H_cons) 
               (proj1_sig St [y:St](member y Q) (delta Hqa))
	       (sig_sig (delta Hqa))).
Apply member_pair_prod.
Assumption.
Exact (monoid_inv_cons_a H_cons).
Save.
Transparent extw.


Lemma extw_def_cons : (w:(word A))(a:A)
  (Ha:(member a I))(Hw:(monoid I w))
  (Hcons:(monoid I (cons A a w)))
  (q:St)(Hq:(member q Q))

  (extw Hcons Hq)
= (extw (monoid_inv_cons_w Hcons) 
        (sig_sig (delta (member_pair_prod Hq (monoid_inv_cons_a Hcons))))).
Proof.
Intros w a Ha Hw Hcons q Hq.
Auto.
Save.

End extw_def.
Transparent extw.
